# For the article see https://www.biorxiv.org/content/10.1101/2020.06.19.159152v1
# For an explenation how to use this layer see https://github.com/ArnovanHilten/GenNet
# Locallyconnected1D is used as a basis to write the LocallyDirected layer
# ==============================================================================
# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""LocallyDirected1D layer.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import tensorflow as tf
from tensorflow.python.keras import activations
from tensorflow.python.keras import backend as K
from tensorflow.python.keras import constraints
from tensorflow.python.keras import initializers
from tensorflow.python.keras import regularizers
from tensorflow.python.keras.engine.base_layer import InputSpec
from tensorflow.python.keras.engine.base_layer import Layer
from tensorflow.python.keras.utils import conv_utils
from tensorflow.python.keras.utils import tf_utils
from tensorflow.python.util.tf_export import tf_export


@tf_export('keras.layers.LocallyDirected1D')
class LocallyDirected1D(Layer):
    """Locally-Directed1D layer for 1D inputs.

  The `LocallyDirected1D` layer works similarly to
  the `Conv1D` layer, except that weights are unshared,
  that is, a different set of filters is applied at each different patch
  of the input.

  Example:
  ```python
      # apply a unshared weight convolution 1d of length 3 to a sequence with
      # 10 timesteps, with 64 output filters
      model = Sequential()
      model.add(LocallyDirected1D(64, 3, input_shape=(10, 32)))
      # now model.output_shape == (None, 8, 64)
      # add a new conv1d on top
      model.add(LocallyDirected1D(32, 3))
      # now model.output_shape == (None, 6, 32)
  ```

  Arguments:
      mask: sparse matrix with shape (input, output) connectivity matrix,
            True defines connection between (in_i, out_j), should be sparse (False,0) >> True
            should be scipy sparese matrix in COO Format!
      filters: Integer, the dimensionality of the output space
          (i.e. the number of output filters in the convolution).
      kernel_size: An integer or tuple/list of a single integer,
          specifying the length of the 1D convolution window.
      strides: An integer or tuple/list of a single integer,
          specifying the stride length of the convolution.
          Specifying any stride value != 1 is incompatible with specifying
          any `dilation_rate` value != 1.
      padding: Currently only supports `"valid"` (case-insensitive).
          `"same"` may be supported in the future.
      data_format: A string,
          one of `channels_last` (default) or `channels_first`.
          The ordering of the dimensions in the inputs.
          `channels_last` corresponds to inputs with shape
          `(batch, length, channels)` while `channels_first`
          corresponds to inputs with shape
          `(batch, channels, length)`.
          It defaults to the `image_data_format` value found in your
          Keras config file at `~/.keras/keras.json`.
          If you never set it, then it will be "channels_last".
      activation: Activation function to use.
          If you don't specify anything, no activation is applied
          (ie. "linear" activation: `a(x) = x`).
      use_bias: Boolean, whether the layer uses a bias vector.
      kernel_initializer: Initializer for the `kernel` weights matrix.
      bias_initializer: Initializer for the bias vector.
      kernel_regularizer: Regularizer function applied to
          the `kernel` weights matrix.
      bias_regularizer: Regularizer function applied to the bias vector.
      activity_regularizer: Regularizer function applied to
          the output of the layer (its "activation")..
      kernel_constraint: Constraint function applied to the kernel matrix.
      bias_constraint: Constraint function applied to the bias vector.

  Input shape:
      3D tensor with shape: `(batch_size, steps, input_dim)`

  Output shape:
      3D tensor with shape: `(batch_size, new_steps, filters)`
      `steps` value might have changed due to padding or strides.
  """

    def __init__(self,
                 mask,
                 filters,
                 padding='valid',
                 data_format=None,
                 activation=None,
                 use_bias=True,
                 kernel_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 **kwargs):
        super(LocallyDirected1D, self).__init__(**kwargs)
        self.filters = filters
        self.padding = conv_utils.normalize_padding(padding)
        self.data_format = conv_utils.normalize_data_format(data_format)
        self.activation = activations.get(activation)
        self.use_bias = use_bias
        self.kernel_initializer = initializers.get(kernel_initializer)
        self.bias_initializer = initializers.get(bias_initializer)
        self.kernel_regularizer = regularizers.get(kernel_regularizer)
        self.bias_regularizer = regularizers.get(bias_regularizer)
        self.activity_regularizer = regularizers.get(activity_regularizer)
        self.kernel_constraint = constraints.get(kernel_constraint)
        self.bias_constraint = constraints.get(bias_constraint)
        self.input_spec = InputSpec(ndim=3)
        self.mask = mask

    @tf_utils.shape_type_conversion
    def build(self, input_shape):
        if self.data_format == 'channels_first':
            input_dim, input_length = input_shape[1], input_shape[2]
        else:
            input_dim, input_length = input_shape[2], input_shape[1]

        if input_dim is None:
            raise ValueError('Axis 2 of input should be fully-defined. '
                             'Found shape:', input_shape)
        self.output_length = self.mask.shape[1]
#         print("output length is " + str(self.output_length))
        if self.data_format == 'channels_first':
            self.kernel_shape = (input_dim, input_length,
                                 self.filters, self.output_length)
        else:
            self.kernel_shape = (input_length, input_dim,
                                 self.output_length, self.filters)

        self.kernel = self.add_weight(shape=(len(self.mask.data),),  # sum of all nonzero values in mask sum(sum(mask))
                                      initializer=self.kernel_initializer,
                                      name='kernel',
                                      regularizer=self.kernel_regularizer,
                                      constraint=self.kernel_constraint)

        self.kernel_mask = get_locallyDirected1D_mask(self.mask, self.kernel,
                                                      data_format=self.data_format,
                                                      dtype=self.kernel.dtype
                                                      )

        if self.use_bias:
            self.bias = self.add_weight(
                shape=(self.output_length, self.filters),
                initializer=self.bias_initializer,
                name='bias',
                regularizer=self.bias_regularizer,
                constraint=self.bias_constraint)
        else:
            self.bias = None

        if self.data_format == 'channels_first':
            self.input_spec = InputSpec(ndim=3, axes={1: input_dim})
        else:
            self.input_spec = InputSpec(ndim=3, axes={-1: input_dim})
        self.built = True

    def call(self, inputs):

        # output = local_conv_matmul(inputs, self.kernel_mask,
        #                            self.output_length)

        output = local_conv_matmul_sparse(inputs, self.kernel_mask,
                                          self.output_length, self.filters)

        if self.use_bias:
            output = K.bias_add(output, self.bias, data_format=self.data_format)

        output = self.activation(output)
        return output

    def get_config(self):  # delete this?
        config = {
            'filters':
                self.filters,
            'padding':
                self.padding,
            'data_format':
                self.data_format,
            'activation':
                activations.serialize(self.activation),
            'use_bias':
                self.use_bias,
            'kernel_initializer':
                initializers.serialize(self.kernel_initializer),
            'bias_initializer':
                initializers.serialize(self.bias_initializer),
            'kernel_regularizer':
                regularizers.serialize(self.kernel_regularizer),
            'bias_regularizer':
                regularizers.serialize(self.bias_regularizer),
            'activity_regularizer':
                regularizers.serialize(self.activity_regularizer),
            'kernel_constraint':
                constraints.serialize(self.kernel_constraint),
            'bias_constraint':
                constraints.serialize(self.bias_constraint),
        }
        base_config = super(LocallyDirected1D, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))


def get_locallyDirected1D_mask(mask, kernel, data_format,
                               dtype):
    """Return a mask representing connectivity of a locally-connected operation.

  This method returns a masking tensor of 0s and 1s (of type `dtype`) that,
  when element-wise multiplied with a fully-connected weight tensor, masks out
  the weights between disconnected input-output pairs and thus implements local
  connectivity through a sparse fully-connected weight tensor.

  Assume an unshared convolution with given parameters is applied to an input
  having N spatial dimensions with `input_shape = (d_in1, ..., d_inN)`
  to produce an output with spatial shape `(d_out1, ..., d_outN)` (determined
  by layer parameters such as `strides`).

  This method returns a mask which can be broadcast-multiplied (element-wise)
  with a 2*(N+1)-D weight matrix (equivalent to a fully-connected layer between
  (N+1)-D activations (N spatial + 1 channel dimensions for input and output)
  to make it perform an unshared convolution with given `kernel_shape`,
  `strides`, `padding` and `data_format`.

  Arguments:
    mask: sparse connectivity matrix matrix
    kernel: weights with len(sum(non-sparse values)
    data_format: a string, `"channels_first"` or `"channels_last"`.
    dtype: type of the layer operation, e.g. `tf.float64`.

  Returns:
    a `dtype`-tensor of shape
    `(1, d_in1, ..., d_inN, 1, d_out1, ..., d_outN)`
    if `data_format == `"channels_first"`, or
    `(d_in1, ..., d_inN, 1, d_out1, ..., d_outN, 1)`
    if `data_format == "channels_last"`.

    Adaption by arno is now a sparse matrix.
  Raises:
    ValueError: if `data_format` is neither `"channels_first"` nor
                `"channels_last"`.
  """

    ndims = int(mask.ndim / 2)
    indices = np.mat([mask.row, mask.col]).transpose()
#     print(mask.shape)
    mask = tf.SparseTensor(indices, kernel, [mask.shape[0], mask.shape[1]])

    if data_format == 'channels_first':
        mask = tf.sparse.expand_dims(mask, 0)
        mask = tf.sparse.expand_dims(mask, - ndims - 1)

    elif data_format == 'channels_last':
        mask = tf.sparse.expand_dims(mask, ndims)
        mask = tf.sparse.expand_dims(mask, -1)

    else:
        raise ValueError('Unrecognized data_format: ' + str(data_format))

    return mask


def local_conv_matmul_sparse(inputs, kernel_mask, output_length, filters):
    """Apply N-D convolution with un-shared weights using a single matmul call.

  This method outputs `inputs . (kernel * kernel_mask)`
  (with `.` standing for matrix-multiply and `*` for element-wise multiply)
  and requires a precomputed `kernel_mask` to zero-out weights in `kernel` and
  hence perform the same operation as a convolution with un-shared
  (the remaining entries in `kernel`) weights. It also does the necessary
  reshapes to make `inputs` and `kernel` 2-D and `output` (N+2)-D.

  Arguments:
      inputs: (N+2)-D tensor with shape
          `(batch_size, channels_in, d_in1, ..., d_inN)`
          or
          `(batch_size, d_in1, ..., d_inN, channels_in)`.
      kernel: the unshared weights for N-D convolution,
          an (N+2)-D tensor of shape:
          `(d_in1, ..., d_inN, channels_in, d_out2, ..., d_outN, channels_out)`
          or
          `(channels_in, d_in1, ..., d_inN, channels_out, d_out2, ..., d_outN)`,
          with the ordering of channels and spatial dimensions matching
          that of the input.
          Each entry is the weight between a particular input and
          output location, similarly to a fully-connected weight matrix.
      kernel_mask: a float 0/1 mask tensor of shape:
           `(d_in1, ..., d_inN, 1, d_out2, ..., d_outN, 1)`
           or
           `(1, d_in1, ..., d_inN, 1, d_out2, ..., d_outN)`,
           with the ordering of singleton and spatial dimensions
           matching that of the input.
           Mask represents the connectivity pattern of the layer and is
           precomputed elsewhere based on layer parameters: stride,
           padding, and the receptive field shape.
      output_shape: a tuple of (N+2) elements representing the output shape:
          `(batch_size, channels_out, d_out1, ..., d_outN)`
          or
          `(batch_size, d_out1, ..., d_outN, channels_out)`,
          with the ordering of channels and spatial dimensions matching that of
          the input.

  Returns:
      Output (N+2)-D tensor with shape `output_shape`.
  """
    inputs_flat = K.reshape(inputs, (K.shape(inputs)[0], -1))
    kernel_mask = make_2d_sparse(kernel_mask, split_dim=K.ndim(kernel_mask) // 2)
    output_flat = tf.sparse.matmul(kernel_mask, inputs_flat, adjoint_a=True, adjoint_b=True)
    output_flat = tf.transpose(output_flat)
    output = K.reshape(output_flat, [-1, output_length, filters])

    return output


def make_2d_sparse(tensor, split_dim):
    """Reshapes an N-dimensional tensor into a 2D tensor.

  Dimensions before (excluding) and after (including) `split_dim` are grouped
  together.

  Arguments:
    tensor: a tensor of shape `(d0, ..., d(N-1))`.
    split_dim: an integer from 1 to N-1, index of the dimension to group
        dimensions before (excluding) and after (including).

  Returns:
    Tensor of shape
    `(d0 * ... * d(split_dim-1), d(split_dim) * ... * d(N-1))`.
  """

    shape = K.array_ops.shape(tensor)
    in_dims = shape[:split_dim]
    out_dims = shape[split_dim:]

    in_size = K.math_ops.reduce_prod(in_dims)
    out_size = K.math_ops.reduce_prod(out_dims)

    return tf.sparse.reshape(tensor, (in_size, out_size))


def local_conv_matmul(inputs, kernel_mask, output_length):
    """Apply N-D convolution with un-shared weights using a single matmul call.

  This method outputs `inputs . (kernel * kernel_mask)`
  (with `.` standing for matrix-multiply and `*` for element-wise multiply)
  and requires a precomputed `kernel_mask` to zero-out weights in `kernel` and
  hence perform the same operation as a convolution with un-shared
  (the remaining entries in `kernel`) weights. It also does the necessary
  reshapes to make `inputs` and `kernel` 2-D and `output` (N+2)-D.

  Arguments:
      inputs: (N+2)-D tensor with shape
          `(batch_size, channels_in, d_in1, ..., d_inN)`
          or
          `(batch_size, d_in1, ..., d_inN, channels_in)`.
      kernel: the unshared weights for N-D convolution,
          an (N+2)-D tensor of shape:
          `(d_in1, ..., d_inN, channels_in, d_out2, ..., d_outN, channels_out)`
          or
          `(channels_in, d_in1, ..., d_inN, channels_out, d_out2, ..., d_outN)`,
          with the ordering of channels and spatial dimensions matching
          that of the input.
          Each entry is the weight between a particular input and
          output location, similarly to a fully-connected weight matrix.
      kernel_mask: a float 0/1 mask tensor of shape:
           `(d_in1, ..., d_inN, 1, d_out2, ..., d_outN, 1)`
           or
           `(1, d_in1, ..., d_inN, 1, d_out2, ..., d_outN)`,
           with the ordering of singleton and spatial dimensions
           matching that of the input.
           Mask represents the connectivity pattern of the layer and is
           precomputed elsewhere based on layer parameters: stride,
           padding, and the receptive field shape.
      output_shape: a tuple of (N+2) elements representing the output shape:
          `(batch_size, channels_out, d_out1, ..., d_outN)`
          or
          `(batch_size, d_out1, ..., d_outN, channels_out)`,
          with the ordering of channels and spatial dimensions matching that of
          the input.

  Returns:
      Output (N+2)-D tensor with shape `output_shape`.
  """
    inputs_flat = K.reshape(inputs, (K.shape(inputs)[0], -1))

    kernel = make_2d_sparse(kernel_mask, split_dim=K.ndim(kernel_mask) // 2)

    output_flat = tf.sparse_tensor_dense_matmul(inputs_flat, kernel, b_is_sparse=True)
    output = K.reshape(output_flat, [-1, output_length, 1])
    return output


def make_2d(tensor, split_dim):
    """Reshapes an N-dimensional tensor into a 2D tensor.

  Dimensions before (excluding) and after (including) `split_dim` are grouped
  together.

  Arguments:
    tensor: a tensor of shape `(d0, ..., d(N-1))`.
    split_dim: an integer from 1 to N-1, index of the dimension to group
        dimensions before (excluding) and after (including).

  Returns:
    Tensor of shape
    `(d0 * ... * d(split_dim-1), d(split_dim) * ... * d(N-1))`.
  """
#     print(tensor.shape)
    shape = K.array_ops.shape(tensor)
    in_dims = shape[:split_dim]
    out_dims = shape[split_dim:]

    in_size = K.math_ops.reduce_prod(in_dims)
    out_size = K.math_ops.reduce_prod(out_dims)

    return K.array_ops.reshape(tensor, (in_size, out_size))

# %%



import os
import sys

sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from gennet.utils.hase.config import *

if PYTHON_PATH is not None:
    for i in PYTHON_PATH: sys.path.insert(0, i)
import pandas as pd
import argparse
from gennet.utils.hase.hdgwas.tools import Reference, Timer
from gennet.utils.hase.hdgwas.hash import *
import gc

if __name__ == '__main__':

    os.environ['HASEDIR'] = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

    parser = argparse.ArgumentParser(description='Script to map studies for meta-stage')
    parser.add_argument("-g", required=True, type=str, help="path/paths to genotype data folder")
    parser.add_argument('-study_name', type=str, required=True, default=None, help=' Study names')
    parser.add_argument("-o", "--out", type=str, required=True, help="path to save result folder")
    parser.add_argument('-ref_name', type=str, default='1000Gp1v3_ref', help='Reference panel name')
    parser.add_argument('-mismatch_table', action='store_true', default=False, help='Save table with mismatch IDs')
    parser.add_argument('-flipped_table', action='store_true', default=False, help='Save table with mismatch IDs')
    parser.add_argument('-probe_chunk', type=int, default=10000, help='Probes chunk')
    parser.add_argument('-ref_chunk', type=int, default=10000, help='Reference chunk')
    parser.add_argument('-chunk', type=int, default=2000000, help='Chunk size')
    args = parser.parse_args()
    print(args)

    try:
        print('Creating directories...')
        os.makedirs(args.out)
    except:
        print(('Directory {} is already exist!'.format(args.out)))

    probes = pd.HDFStore(os.path.join(args.g, 'probes', args.study_name + '.h5'), 'r')
    probes_n_rows = probes.get_storer('probes').nrows
    chunk_size = np.min([args.chunk, probes_n_rows])

    print(('Merge chunk size {}'.format(chunk_size)))
    match_key = np.array([], dtype=np.int32)
    match_index = np.array([], dtype=np.int32)
    flip_key = np.array([], dtype=np.int32)
    flip_index = np.array([], dtype=np.int32)
    ID = np.array([])

    del_counter_ref = {}
    ID2CHR = False
    IDconv = False
    hashing = False
    merge_on = {
        'ID': {
            'straight': ["ID", 'allele1', 'allele2'],
            'reverse': ["ID", 'allele2', 'allele1']
        },

        'CHR': {
            'straight': ["CHR", 'bp', 'allele1', 'allele2'],
            'reverse': ["CHR", 'bp', 'allele2', 'allele1']

        }
    }

    for p in range(int(np.ceil(probes_n_rows / float(chunk_size)))):
        print('p', p)

        p_start_i = p * chunk_size
        p_stop_i = min((p + 1) * chunk_size, probes_n_rows)

        a = probes.select('probes', start=p_start_i, stop=p_stop_i)

        if p == 0:
            print(a.head())
            if issubclass(type(a.iloc[0]['allele1']), np.str):
                hashing = True
            if "CHR" in a.columns and 'bp' in a.columns:
                ID2CHR = True
                merge = merge_on['CHR']
                print('Merge on CHR/bp')
            else:
                if ':' in a.ID.iloc[0] and ':' in a.ID.iloc[1]:
                    CHR = []
                    bp = []
                    for i in a.ID:
                        s = i.split(":")
                        CHR.append(s[0])
                        bp.append(s[1])
                    CHR = np.array(CHR, dtype=np.int8)
                    bp = np.array(bp)
                    if np.max(CHR) < 23 and np.min(CHR) > 0:
                        a['CHR'] = CHR
                        a['bp'] = bp
                        a.CHR = a.CHR.astype(np.int64)
                        a.bp = a.bp.astype(np.int64)
                        ID2CHR = True
                        IDconv = True
                        merge = merge_on['CHR']
                        print('Merge on CHR/bp from ID')
                        print(a.head())
                    else:
                        print('No CHR and bp info...')
                        merge = merge_on['ID']
                        print('Merge on ID')
                else:
                    print('No CHR and bp info...')
                    merge = merge_on['ID']
                    print('Merge on ID')

        elif IDconv:
            def f(x):
                s = x.ID.split(':')
                return s[0], s[1]


            CHR_bp = a.apply(f, axis=1)
            a['CHR'], a['bp'] = list(zip(*CHR_bp))
            a.CHR = a.CHR.astype(np.int64)
            a.bp = a.bp.astype(np.int64)
            print(a.head())
        a['counter_prob'] = np.arange(p_start_i, p_stop_i, dtype='int32')

        reference = Reference()
        reference.name = args.ref_name
        reference.chunk = args.ref_chunk
        reference.load()
        counter_ref = 0
        if hashing:
            print('Hashing...')
            a.allele1 = a.allele1.apply(hash)
            a.allele2 = a.allele2.apply(hash)
        for r, b in enumerate(reference.dataframe):
            if r == 0:
                if np.sum(np.array([1 if i in reference.columns else 0 for i in b.columns.tolist()])) != len(
                        reference.columns):
                    raise ValueError('Reference table should have {} columns'.format(reference.columns))
            if r == 0 and p == 0:
                print('********************************')
                print(('Use {} as a reference panel'.format(args.ref_name)))
                print(b.head())
                print('********************************')

            print('r', r)
            if p == 0:
                ID = np.append(ID, b.ID)

            b['counter_ref'] = np.arange(counter_ref, counter_ref + b.shape[0], dtype='int32')
            counter_ref += b.shape[0]

            if len(match_index) or len(flip_index):
                print('matched {}'.format(match_index.shape[0]))
                print('flipped {}'.format(flip_index.shape[0]))
                if del_counter_ref.get(r) is not None:
                    with Timer() as t:
                        b = b[~b.counter_ref.isin(del_counter_ref[r])]
                    print('time {}'.format(t.secs))

            match_df = pd.merge(b, a, left_on=merge['straight'], right_on=merge['straight'])
            flip_df = pd.merge(b[~b.counter_ref.isin(match_df.counter_ref)], a, left_on=merge['reverse'],
                               right_on=merge['straight'])

            if len(match_df):
                match_key = np.append(match_key, match_df.counter_ref)
                match_index = np.append(match_index, match_df.counter_prob)
                if del_counter_ref.get(r) is None:
                    del_counter_ref[r] = match_key
                else:
                    del_counter_ref[r] = np.append(del_counter_ref[r], match_key)
            if len(flip_df):
                flip_key = np.append(flip_key, flip_df.counter_ref)
                flip_index = np.append(flip_index, flip_df.counter_prob)
                if del_counter_ref.get(r) is None:
                    del_counter_ref[r] = flip_key
                else:
                    del_counter_ref[r] = np.append(del_counter_ref[r], flip_key)
            gc.collect()

    index = np.ones(ID.shape[0], dtype='int') * -1
    flip = np.ones(probes_n_rows, dtype='int')
    index[match_key] = match_index
    index[flip_key] = flip_index
    flip[flip_index] = -1
    print(('Saving results for {} to {} ...'.format(args.study_name, args.out)))
    np.save(os.path.join(args.out, 'values_' + reference.name + '_' + args.study_name + '.npy'), index)
    np.save(os.path.join(args.out, 'flip_' + reference.name + '_' + args.study_name + '.npy'), flip)
    np.save(os.path.join(args.out, 'keys_' + reference.name + '.npy'), ID)
    print('Data successfully saved')

    mismatch_index = np.setdiff1d(np.arange(probes_n_rows), np.append(match_index, flip_index))

    if os.path.isfile(os.path.join(args.g, 'probes', args.study_name + '_hash_table.csv.gz')):
        try:
            df_hash = pd.read_csv(os.path.join(args.g, 'probes', args.study_name + '_hash_table.csv.gz'), sep='\t',
                                  compression='gzip', index_col=False)
        except:
            df_hash = pd.read_csv(os.path.join(args.g, 'probes', args.study_name + '_hash_table.csv.gz'), sep='\t',
                                  index_col=False)


    else:
        df_hash = None
        print(('You do not have hash_table for alleles in your probes folder! '
               'You used old version of HASE to convert your genotype data.'
               'To see original codes for allele you can make hash_table using script'
               '{}/tools/tools.py -hash -g "original genotype folder" '.format(os.environ['HASEDIR'])))

    print('There are {} common variances with reference panel, which will be included in study'.format(
        np.where(index != -1)[0].shape[0]))
    print('There are {} variances from reference panel, which were not found in probes'.format(
        np.where(index == -1)[0].shape[0]))
    print('There are {} variances excluded from study (not found in reference panel)'.format(
        probes_n_rows - np.where(index != -1)[0].shape[0]))
    if args.mismatch_table and mismatch_index.shape[0] != 0:
        df_mismatch = probes.select('probes', where=mismatch_index)
        if df_hash is not None and not hashing:
            df_mismatch = pd.merge(df_hash, df_mismatch, left_on='keys', right_on='allele1')
            df_mismatch['str_allele1'] = df_mismatch['allele']
            del df_mismatch['allele']
            df_mismatch = pd.merge(df_hash, df_mismatch, left_on='keys', right_on='allele2')
            df_mismatch['str_allele2'] = df_mismatch['allele']
            del df_mismatch['allele']
            del df_mismatch['keys_x']
            del df_mismatch['keys_y']
        df_mismatch.to_csv(os.path.join(args.out, 'mismatch_ID_info.csv'))
        print('Mismatch ID info saved to {}'.format(os.path.join(args.out, args.study_name + '_mismatch_ID_info.csv')))
    elif mismatch_index.shape[0] != 0:
        print('Mismatch examples:')
        print(probes.select('probes', where=mismatch_index[:10]))

    print('There are {} flipped variances'.format(len(flip_index)))
    if args.flipped_table and flip_index.shape[0] != 0:
        df_flipped = probes.select('probes', where=flip_index)
        if df_hash is not None and not hashing:
            df_flipped = pd.merge(df_hash, df_flipped, left_on='keys', right_on='allele1')
            df_flipped['str_allele1'] = df_flipped['allele']
            del df_flipped['allele']
            df_flipped = pd.merge(df_hash, df_flipped, left_on='keys', right_on='allele2')
            df_flipped['str_allele2'] = df_flipped['allele']
            del df_flipped['allele']
            del df_flipped['keys_x']
            del df_flipped['keys_y']
        df_flipped.to_csv(os.path.join(args.out, 'flipped_ID_info.csv'))
        print('Flipped ID info saved to {}'.format(os.path.join(args.out, args.study_name + '_flipped_ID_info.csv')))
    elif flip_index.shape[0] != 0:
        print('Flipped examples:')
        print(probes.select('probes', where=flip_index[:10]))



import os
import numpy as np
import pandas as pd
import scipy
import scipy.sparse


def Create_Annovar_input(args):
    hasepath = args.path
    studyname = args.study_name
    savepath = args.out

    if os.path.exists(hasepath + '/probes/' + studyname + '_selected.h5'):
        probes = pd.read_hdf(hasepath + '/probes/' + studyname + '_selected.h5', mode="r")
    else:
        probes = pd.read_hdf(hasepath + '/probes/' + studyname + '.h5', mode="r")
        print(probes.shape)

    if os.path.exists(hasepath + '/probes/' + studyname + '_hash_table.csv.gz'):
        hashtable = pd.read_csv(hasepath + '/probes/' + studyname + '_hash_table.csv.gz', compression="gzip", sep='\t')
    else:
        hashtable = pd.read_csv(hasepath + '/probes/' + studyname + '_hash_table.csv', sep='\t')

    hashtable['allele1'] = hashtable['keys']
    unhashed_probes = probes.merge(hashtable, on='allele1', how="left")
    unhashed_probes = unhashed_probes.drop(columns=["keys", "allele1"])
    unhashed_probes = unhashed_probes.rename(columns={'allele': 'allele1'})

    # reload hashtable for other allele

    if os.path.exists(hasepath + '/probes/' + studyname + '_hash_table.csv.gz'):
        hashtable = pd.read_csv(hasepath + '/probes/' + studyname + '_hash_table.csv.gz', compression="gzip", sep='\t')
    else:
        hashtable = pd.read_csv(hasepath + '/probes/' + studyname + '_hash_table.csv', sep='\t')

    hashtable['allele2'] = hashtable['keys']
    unhashed_probes = unhashed_probes.merge(hashtable, on='allele2', how="left")
    unhashed_probes = unhashed_probes.drop(columns=["keys", "allele2"])
    unhashed_probes = unhashed_probes.rename(columns={'allele': 'allele2'})

    # clean
    annovar_input = unhashed_probes.drop(columns=["ID", "distance"])
    annovar_input["bp2"] = annovar_input["bp"]
    annovar_input["index_col"] = annovar_input.index
    annovar_input = annovar_input[['CHR', 'bp', "bp2", "allele1", "allele2", "index_col"]]

    print('Number of variants', annovar_input.shape)

    annovar_input_path = savepath + '/annovar_input_' + studyname + '.csv'
    annovar_input.to_csv(annovar_input_path, sep="\t", index=False, header=False)

    print('\n')
    print('Annovar input files ready \n')
    print("Install annovar: https://annovar.openbioinformatics.org/en/latest/user-guide/download/")
    print("Navigate to annovar, e.g cd /home/charlesdarwin/annovar/")
    print("Update annovar:\n perl annotate_variation.pl -buildver hg19 -downdb -webfrom annovar refGene humandb/")
    print("Run:\n perl annotate_variation.pl -geneanno -dbtype refGene -buildver hg19 " + str(
        savepath) + "/annovar_input_" + str(studyname) + ".csv humandb --outfile " + str(savepath) + "/" + str(
        studyname) + "_RefGene")
    print('\n')
    print(
        'After obtaining the Annovar annotations, run topology create_gene_network to get the topology file for the SNPs-gene-output network:')


def Create_gene_network_topology(args):
    datapath = args.path + '/'
    studyname = args.study_name
    savepath = args.out + '/'

    print(args.study_name)

    gene_annotation = pd.read_csv(datapath + str(studyname) + "_RefGene.variant_function", sep='\t', header=None)
    gene_annotation.columns = ['into/exonic', 'gene', 'chr', 'bps', 'bpe', "mutation1", "mutation2", 'index_col']
    gene_annotation['gene'] = gene_annotation['gene'].str.replace(r"\,.*", "")
    # gene_annotation['dist'] = gene_annotation['gene'].str.extract(r"(?<=dist\=)(.*)(?=\))")
    gene_annotation['gene'] = gene_annotation['gene'].str.replace(r"\(.*\)", "")
    gene_annotation['gene'] = gene_annotation['gene'].str.replace(r"\(.*", "")
    gene_annotation['gene'] = gene_annotation['gene'].str.replace(r"\;.*", "")
    gene_annotation = gene_annotation[(gene_annotation['gene'] != "NONE")]
    gene_annotation = gene_annotation.dropna()

    gene_list = gene_annotation.drop_duplicates("gene")
    gene_list = gene_list.sort_values(by=["chr", "bps"], ascending=[True, True])
    gene_list["gene_id"] = np.arange(len(gene_list))
    gene_list = gene_list[["gene", "gene_id"]]

    gene_annotation = gene_annotation.merge(gene_list, on="gene")
    gene_annotation = gene_annotation.sort_values(by="index_col", ascending=True)

    gene_annotation = gene_annotation.assign(
        chrbp='chr' + gene_annotation.chr.astype(str) + ':' + gene_annotation.bps.astype(str))
    gene_annotation.to_csv(savepath + "/gene_network_description.csv")

    topology = gene_annotation[["chr", "index_col", "chrbp", "gene_id", "gene"]]
    print(topology['index_col'].max())
    topology.columns = ['chr', 'layer0_node', 'layer0_name', 'layer1_node', 'layer1_name']

    topology.to_csv(savepath + "/topology.csv")

    print('Topology file saved:', savepath + "/topology.csv")

    # Additionally create a mask for creating larger networks
    data = np.ones(gene_annotation.shape[0], bool)
    coord = (gene_annotation["index_col"].values, gene_annotation["gene_id"].values)
    SNP_gene_matrix = scipy.sparse.coo_matrix(((data), coord),
                                              shape=(gene_annotation["index_col"].max() + 1,
                                                     gene_annotation["gene_id"].max() + 1))
    scipy.sparse.save_npz(savepath + '/SNP_gene_mask', SNP_gene_matrix)
    print("Alternatively you can choose to use the .npz mask (building blocks for deeper networks)",
          savepath + '/SNP_gene_mask', 'The mask has shape:', SNP_gene_matrix.shape)

    return gene_annotation


def Create_gene_to_pathway_KEGG(args):
    gene_overview = Create_gene_network_topology(args)
    savepath = args.out + '/'

    pathway_overview_higher_levels = pd.read_csv('resources/pathways/pathway_overview_KEGG.csv').drop(
        "local_id", axis=1)

    CPDB_pathway_overview = pd.read_csv('resources/pathways/CPDB_pathways_genes.tab', sep='\t')

    pathway_overview_source = CPDB_pathway_overview[CPDB_pathway_overview["source"] == "KEGG"].copy()
    pathway_overview_source["hsaid"] = pd.to_numeric(pathway_overview_source['external_id'].str.replace("path:hsa", ""),
                                                     errors='coerce')
    pathway_overview_source = pathway_overview_source.merge(pathway_overview_higher_levels, on="hsaid")
    pathway_overview_source["local_id"] = np.arange(len(pathway_overview_source))
    pathway_overview_source["num_genes"] = pathway_overview_source["hgnc_symbol_ids"].str.split(",").str.len()

    coordinate = np.empty(shape=(0, 2), dtype=int)
    for pathnum in pathway_overview_source["local_id"]:
        gene_ids_in_pathway = np.array(gene_overview[gene_overview["gene"].isin(
            set(pathway_overview_source["hgnc_symbol_ids"].iloc[pathnum].split(",")))]["gene_id"], dtype=int)
        current_coordinates = np.ones((len(gene_ids_in_pathway), 2), dtype=int) * pathnum
        current_coordinates[:, 0] = gene_ids_in_pathway
        # print(pathnum, current_coordinates.shape[0], pathway_overview_source["num_genes"].iloc[pathnum])
        coordinate = np.append(coordinate, current_coordinates, axis=0)

    data = np.ones(len(coordinate), np.bool)
    mask_gene_local = scipy.sparse.coo_matrix(((data), (coordinate[:, 0], coordinate[:, 1])), shape=(
        gene_overview["gene_id"].max() + 1, pathway_overview_source["local_id"].max() + 1))
    scipy.sparse.save_npz(savepath + 'mask_gene_local', mask_gene_local)
    print("mask_gene_local saved in ", savepath, "with shape", mask_gene_local.shape, "and", len(coordinate),
          "elements")

    HSA_overview_mid_unique = pathway_overview_source.drop_duplicates("local_id")

    mask_pathway_mid = scipy.sparse.coo_matrix(
        (np.ones(len(HSA_overview_mid_unique), np.bool),
         (HSA_overview_mid_unique['local_id'], HSA_overview_mid_unique['mid_id'].values)),
        shape=(HSA_overview_mid_unique['local_id'].max() + 1, HSA_overview_mid_unique['mid_id'].max() + 1),
    )
    scipy.sparse.save_npz(savepath + 'mask_local_mid', mask_pathway_mid)
    print("mask_local_mid saved in ", savepath, "with shape", mask_pathway_mid.shape, "and",
          len(HSA_overview_mid_unique), "elements")

    HSA_overview_global_unique = pathway_overview_source.drop_duplicates("mid_id")

    mask_pathway_global = scipy.sparse.coo_matrix(
        (np.ones(len(HSA_overview_global_unique), np.bool),
         (HSA_overview_global_unique['mid_id'], HSA_overview_global_unique['global_id'].values)),
        shape=(HSA_overview_global_unique['mid_id'].max() + 1, HSA_overview_global_unique['global_id'].max() + 1),
    )
    scipy.sparse.save_npz(savepath + 'mask_mid_global', mask_pathway_global)
    print("mask_mid_global saved in ", savepath, "with shape", mask_pathway_global.shape,
          "and", len(HSA_overview_global_unique), "elements")


def Create_gene_to_GTEx(args):
    print("This will be implemented in the next version")
    raise NotImplementedError


def topology(args):
    if args.type == 'create_annovar_input':
        Create_Annovar_input(args)
    elif args.type == 'create_gene_network':
        Create_gene_network_topology(args)
    elif args.type == 'create_pathway_KEGG':
        Create_gene_to_pathway_KEGG(args)
    elif args.type == 'create_GTEx_network':
        Create_gene_to_GTEx(args)
    else:
        print("invalid type:", args.type)
        exit()



from setuptools import find_packages, setup

setup(
    package_dir={"gennet": "gennet"},
    packages=find_packages(
        where=".",
        include=[
            "*",
        ],
        exclude=["tests", "tests.*"],
    ),
)



import os
import sys

sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from gennet.utils.hase.config import basedir, PYTHON_PATH

os.environ['HASEDIR'] = basedir
if PYTHON_PATH is not None:
    for i in PYTHON_PATH: sys.path.insert(0, i)
from gennet.utils.hase.hdgwas.tools import HaseAnalyser
import argparse
import pandas as pd
import numpy as np
from collections import OrderedDict

if __name__ == "__main__":

    os.environ['HASEDIR'] = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    parser = argparse.ArgumentParser(description='Script analyse results of HASE')
    parser.add_argument("-r", required=True, help="path to hase results")
    parser.add_argument("-o", "--out", type=str, required=True, help="path to save result folder")
    parser.add_argument("-df", type=float, default=None,
                        help="degree of freedom = ( #subjects in study  - #covariates - 1 )")
    parser.add_argument("-N", type=int, default=None, help="file number to read")
    # TODO (low) add reference panel
    args = parser.parse_args()
    Analyser = HaseAnalyser()
    print(args)

    Analyser.DF = args.df
    Analyser.result_path = args.r
    Analyser.file_number = args.N

    results = OrderedDict()
    results['RSID'] = np.array([])
    results['p_value'] = np.array([])
    results['t-stat'] = np.array([])
    results['phenotype'] = np.array([])
    results['SE'] = np.array([])
    results['MAF'] = np.array([])
    results['BETA'] = np.array([])

    while True:
        Analyser.summary()
        if Analyser.results is None:
            break
        print('Saving data...')
        if not os.path.exists(os.path.join(args.out, 'results' + '.csv')):
            df = pd.DataFrame.from_dict(results)
            df.to_csv(os.path.join(args.out, 'results' + '.csv'), sep=" ", index=None)
        df = pd.DataFrame.from_dict(Analyser.results)
        with open(os.path.join(args.out, 'results' + '.csv'), 'a') as f:
            df.to_csv(f, sep=" ", header=False, index=None)



import os
import time

import numpy as np

from gennet.utils.hase.hdgwas.hdregression import A_covariates, A_tests, B_covariates, C_matrix, B4
from gennet.utils.hase.hdgwas.tools import study_indexes, Timer


def merge_PD(path, max_node, study_name):
    print('Merging PD...')
    while True:
        time.sleep(10)
        if np.sum([os.path.isfile(os.path.join(path, 'node_{}_{}_metadata.npy'.format(i, study_name))) for i in
                   range(1, max_node + 1)]) == max_node:
            if np.sum([os.path.isfile(os.path.join(path, 'node_{}_{}_a_test.npy'.format(i, study_name))) for i in
                       range(1, max_node + 1)]) == max_node:
                break
    print('All files found...')
    b4_flag = os.path.isfile(os.path.join(path, 'node_{}_{}_b4.npy'.format(1, study_name)))
    for i in range(1, max_node + 1):
        print('node_{}'.format(i))
        if i == 1:
            metadata = np.load(os.path.join(path, 'node_{}_{}_metadata.npy'.format(i, study_name))).item()
            a_test = np.load(os.path.join(path, 'node_{}_{}_a_test.npy'.format(i, study_name)))
            if b4_flag:
                b4 = np.load(os.path.join(path, 'node_{}_{}_b4.npy'.format(i, study_name)))


        else:
            metadata_tmp = np.load((os.path.join(path, 'node_{}_{}_metadata.npy'.format(i, study_name)))).item()
            a_test_tmp = np.load((os.path.join(path, 'node_{}_{}_a_test.npy'.format(i, study_name))))
            metadata['MAF'] = metadata['MAF'] + metadata_tmp['MAF']
            a_test = np.vstack((a_test, a_test_tmp))
            if b4_flag:
                b4_tmp = np.load((os.path.join(path, 'node_{}_{}_b4.npy'.format(i, study_name))))
                b4 = np.vstack((b4, b4_tmp))

    np.save(os.path.join(path, study_name + '_a_test.npy'), a_test)
    np.save(os.path.join(path, study_name + '_metadata.npy'), metadata)
    if b4_flag:
        np.save(os.path.join(path, study_name + '_b4.npy'), b4)

    for i in range(1, max_node + 1):
        os.remove(os.path.join(path, 'node_{}_{}_metadata.npy'.format(i, study_name)))
        os.remove(os.path.join(path, 'node_{}_{}_a_test.npy'.format(i, study_name)))
        if b4_flag:
            os.remove(os.path.join(path, 'node_{}_{}_b4.npy'.format(i, study_name)))


def partial_derivatives(save_path=None, COV=None, PHEN=None, GEN=None,
                        MAP=None, MAF=None, R2=None, B4_flag=False, study_name=None, intercept=True):
    row_index, ids = study_indexes(phenotype=PHEN.folder._data, genotype=GEN.folder._data, covariates=COV.folder._data)

    metadata = {}

    # TODO (mid) add parameter to compute PD only for new phenotypes or cov
    metadata['id'] = ids
    metadata['MAF'] = []
    metadata['filter'] = []
    metadata['names'] = []  # TODO (low) change to cov_names
    metadata['phenotype'] = []
    b_cov = []
    C = []
    a_test = []
    b4 = []

    covariates = COV.get_next(index=row_index[2])

    if MAP.cluster == 'n' or MAP.node[1] == 1:
        if intercept:
            metadata['names'].append(study_name + '_intercept')
        metadata['names'] = metadata['names'] + [study_name + '_' + i for i in COV.folder._data.get_names()]

        a_cov = A_covariates(covariates, intercept=intercept)
        np.save(os.path.join(save_path, study_name + '_a_cov.npy'), a_cov)

        with Timer() as t_phen:

            while True:

                phenotype = PHEN.get_next(index=row_index[1])
                if isinstance(phenotype, type(None)):
                    b_cov = np.concatenate(b_cov, axis=1)
                    C = np.concatenate(C, axis=0)
                    np.save(os.path.join(save_path, study_name + '_b_cov.npy'), b_cov)
                    np.save(os.path.join(save_path, study_name + '_C.npy'), C)
                    break

                metadata['phenotype'] = metadata['phenotype'] + list(PHEN.folder._data.get_names())
                b_cov.append(B_covariates(covariates, phenotype, intercept=intercept))
                C.append(C_matrix(phenotype))

        print(('Time to PD phenotype {} is {} s'.format(np.array(C).shape, t_phen.secs)))

    if MAP.cluster == 'y':
        f_max = np.max([int(f.split('_')[0]) for f in GEN.folder.files])
        files2read = ['{}_{}.h5'.format(i, study_name) for i in
                      np.array_split(list(range(f_max + 1)), MAP.node[0])[MAP.node[1] - 1]][::-1]
        filesdone = []
        for i in range(MAP.node[1] - 1):
            filesdone = filesdone + ['{}_{}.h5'.format(i, study_name) for i in
                                     np.array_split(list(range(f_max + 1)), MAP.node[0])[i]]

        N_snps_read = 0
        for f in filesdone:
            file = os.path.join(GEN.folder.path, 'genotype', f)
            N_snps_read += GEN.folder.get_info(file)['shape'][0]
    else:
        N_snps_read = 0
    while True:
        with Timer() as t_gen:
            if MAP.cluster == 'y':
                if len(files2read) != 0:
                    file = os.path.join(GEN.folder.path, 'genotype', files2read.pop())
                    genotype = GEN.folder.read(file)
                else:
                    genotype = None
            else:
                genotype = GEN.get_next()
            if isinstance(genotype, type(None)):
                if MAP.cluster == 'y':

                    np.save(os.path.join(save_path, 'node_{}_'.format(MAP.node[1]) + study_name + '_a_test.npy'),
                            np.concatenate(a_test).astype(np.float64))
                    np.save(os.path.join(save_path, 'node_{}_'.format(MAP.node[1]) + study_name + '_metadata.npy'),
                            metadata)
                    if B4_flag:
                        b4 = np.concatenate(b4, axis=0)
                        np.save(os.path.join(save_path, 'node_{}_'.format(MAP.node[1]) + study_name + '_b4.npy'),
                                b4.astype(np.float64))
                    if MAP.node[1] == MAP.node[0]:
                        merge_PD(save_path, MAP.node[0], study_name)

                else:
                    np.save(os.path.join(save_path, study_name + '_a_test.npy'), np.concatenate(a_test))
                    np.save(os.path.join(save_path, study_name + '_metadata.npy'), metadata)
                    if B4_flag:
                        b4 = np.concatenate(b4, axis=0)
                        np.save(os.path.join(save_path, study_name + '_b4.npy'), b4)
                break
            flip = MAP.flip[GEN.folder.name][N_snps_read:N_snps_read + genotype.shape[0]]
            N_snps_read += genotype.shape[0]
            flip_index = (flip == -1)
            genotype = np.apply_along_axis(lambda x: flip * (x - 2 * flip_index), 0, genotype)
            genotype = genotype[:, row_index[0]]
            maf = np.mean(genotype, axis=1) / 2
            metadata['MAF'] = metadata['MAF'] + list(maf)

            # TODO (low) add interaction
            a_test.append(A_tests(covariates, genotype, intercept=intercept))

            if B4_flag:
                # works only when all phenotypes in one chunk, if not, do not use this option!
                # it would use to much disk space anyway
                if len([f for f in PHEN.folder.files if f != 'info_dic.npy']) > 1:
                    print('pd_full flag disabled!')
                    B4_flag = False
                    continue
                PHEN.folder.processed = 0
                phenotype = PHEN.get_next(index=row_index[1])
                b4.append(B4(phenotype, genotype))

        print(('Time to PD genotype {} is {} s'.format(genotype.shape, t_gen.secs)))



import os
import sys
import warnings

warnings.filterwarnings('ignore')
import argparse


def main():
    args = ArgumentParser().parse_cmd_args()

    if "out" in args and not os.path.exists(args.out):
            print("-out directory did not exist but is made now")
            os.makedirs(args.out)


    if args.mode == 'train':
        if args.problem_type == "classification":
            args.regression = False
        elif args.problem_type == "regression":
            args.regression = True
        else:
            print('something went wrong invalid problem type', args.problem_type)
        from gennet.utils.Train_network import train_model
        train_model(args)
        
    elif args.mode == "plot":
        from gennet.utils.Create_plots import plot
        plot(args)
    if args.mode == 'convert':
        from gennet.utils.Convert import convert
        convert(args)
    if args.mode == "topology":
        from gennet.utils.Topology import topology
        topology(args)
    if args.mode == "interpret":
        from gennet.utils.Interpret import interpret
        interpret(args)


class ArgumentParser():
    """Argumentparser"""

    def __init__(self):
        parser = argparse.ArgumentParser(description="GenNet: Interpretable neural networks for phenotype prediction.",
                                         epilog="Check the wiki on github.com/arnovanhilten/gennet/ for more info")

        subparsers = parser.add_subparsers(help="GenNet main options", dest="mode")

        parser_convert = subparsers.add_parser("convert", help="Convert genotype data to hdf5")
        self.make_parser_covert(parser_convert)

        parser_train = subparsers.add_parser("train", help="Trains the network")
        self.make_parser_train(parser_train)

        parser_plot = subparsers.add_parser("plot", help="Generate plots from a trained network")
        self.make_parser_plot(parser_plot)

        parser_topology = subparsers.add_parser("topology", help="Create standard topology files")
        self.make_parser_topology(parser_topology)

        parser_interpret = subparsers.add_parser("interpret", help="Post-hoc interpretation analysis on the network")
        self.make_parser_interpret(parser_interpret)

        self.parser = parser

    def parse_cmd_args(self):
        args = self.parser.parse_args()
        return args

    def make_parser_covert(self, parser_convert):
        parser_convert.add_argument(
            "-g", "--genotype",
            nargs='+',
            type=str,
            help="Path/paths to genotype data folder")
        parser_convert.add_argument(
            '-study_name',
            type=str,
            required=True,
            nargs='+',
            help=' Name for saved genotype data, without ext')
        parser_convert.add_argument(
            '-variants',
            type=str,
            help="Path to file with row numbers of variants to include, if none is "
                 "given all variants will be used",
            default=None)
        parser_convert.add_argument(
            "-o", "--out",
            type=str,
            default=os.getcwd() + '/processed_data/',
            help="Path for saving the results, default ./processed_data")
        parser_convert.add_argument(
            '-ID',
            action='store_true',
            default=False,
            help='Flag to convert minimac data to genotype per subject files first (default '
                 'False)')
        parser_convert.add_argument(
            '-vcf',
            action='store_true',
            default=False,
            help='Flag for VCF data to convert')
        parser_convert.add_argument(
            '-tcm',
            type=int,
            default=500000000,
            help='Modifier for chunk size during TRANSPOSING make it lower if you run out of '
                 'memory during transposing')
        parser_convert.add_argument(
            '-step',
            type=str,
            default='all',
            choices=['all', 'hase_convert', 'merge', 'impute_missing', 'exclude', 'transpose',
                     'merge_transpose', 'checksum'],
            help='Modifier to choose step to do')
        parser_convert.add_argument(
            '-n_jobs',
            type=int,
            default=1,
            help='Choose jobs > 1 for multiple job submission on a cluster')
        parser_convert.add_argument(
            '-comp_level',
            type=int,
            default=1,
            help='How compressed should the data be? Between 1-9. 1 \
                                    for low compression, 9 is highest compression')
        return parser_convert

    def make_parser_train(self, parser_train):
        parser_train.add_argument(
            "-path",
            type=str,
            help="Path to the data. Subject file, npz masks/topology and/or genotype.h5",
            required=True)
        parser_train.add_argument(
            "-ID",
            type=int,
            help="Number of the experiment",
            required=True)
        parser_train.add_argument(
            "-genotype_path",
            type=str,
            help="Path to genotype data if the location is not the same as given in -path",
            default="undefined")
        parser_train.add_argument(
            "-problem_type",
            default='classification', type=str,
            choices=['classification', 'regression'],
            help="Type of problem, choices are: classification or regression")
        parser_train.add_argument(
            "-wpc",
            type=float,
            metavar="weight positive class",
            default=1,
            help="Hyperparameter:weight of the positive class")
        parser_train.add_argument(
            "-lr", '--learning_rate',
            type=float,
            metavar="learning rate",
            default=0.001,
            help="Hyperparameter: learning rate of the optimizer")
        parser_train.add_argument(
            "-bs", '--batch_size',
            type=int,
            metavar="batch size",
            default=32,
            help='Hyperparameter: batch size')
        parser_train.add_argument(
            "-epochs",
            type=int,
            metavar="number of epochs",
            default=1000,
            help='Hyperparameter: batch size')
        parser_train.add_argument(
            "-workers",
            type=int,
            metavar="number of workers for multiprocessing",
            default=1,
            help='Speed-up: number of workers (CPU cores) for multiprocessing. Can cause memory-leaks in some tensorflow versions')
        parser_train.add_argument(
            "-L1",
            metavar="",
            type=float,
            default=0.01,
            help='Hyperparameter: value for the L1 regularization pentalty similar as in lasso, enforces sparsity')
        parser_train.add_argument(
            "-L1_act",
            metavar="",
            type=float,
            default=0.01,
            help='Hyperparameter: value for the L1 regularization on the activation, enforces sparse activations')
        parser_train.add_argument(
            "-network_name",
            type=str,
            help="Name of the network",
            default="undefined")
        parser_train.add_argument(
            "-filters",
            type=int,
            metavar="number of filters for the gene layer",
            default=2,
            help='Hyperparameter: number of filters for the gene layer')
        parser_train.add_argument(
            "-mixed_precision",
            action='store_true',
            default=False,
            help='Flag for mixed precision to save memory (can reduce performance)')
        parser_train.add_argument(
            "-suffix",
            metavar="extra_info",
            type=str,
            default='',
            help='Add extra suffix for easier identification of the folder')
        parser_train.add_argument(
            "-out",
            metavar="outfolder",
            type=str,
            default='undefined',
            help='Use this argument to change the output directory')
        parser_train.add_argument(
            "-mask_order",
            metavar="mask_order",
            nargs='+',
            default=[],
            help='Use this to define the order of the mask if they should not be ordered by size. '
                 'list masks by full name and in order. (e.g. --mask_order SNP_gene_mask mask_gene_local'
                 ' mask_local_mid mask_mid_global)')
        parser_train.add_argument(
            "-epoch_size",
            metavar="epoch_size",
            type=int,
            default=None,
            help='Use this argument to shorten an epoch if an epoch takes to long.'
                 'Epoch_size will be the new epoch size. Epochs will be shuffled after all data has been seen')
        parser_train.add_argument(
            "-patience",
            metavar="patience",
            type=int,
            default=50,
            help='Number of epochs with no improvement after which training will be stopped.')
        parser_train.add_argument(
            "-resume",
            action='store_true',
            default=False,
            help='Flag for resuming training with existing weights (if they exist)')
        parser_train.add_argument(
            "-onehot",
            action='store_true',
            default=False,
            help='Flag for one hot encoding as a first layer in the network')        
        parser_train.add_argument(
            "-init_linear",
            action='store_true',
            default=False,
            help='initialize the one-hot encoding for the neural network with a linear assumption')
        return parser_train

    def make_parser_plot(self, parser_plot):
        parser_plot.add_argument(
            "-ID",
            type=int,
            help="ID of the experiment",
            required=True)
        parser_plot.add_argument(
            "-type",
            type=str,
            choices=['layer_weight', 'sunburst', 'manhattan_relative_importance'],
            required=True)
        parser_plot.add_argument(
            "-layer_n",
            type=int,
            help="Only used for layer weight: Number of the to be plotted layer",
            metavar="Layer_number:",
            default=0)
        parser_plot.add_argument(
            "-out",
            metavar="outfolder",
            type=str,
            default='undefined',
            help='Use this argument to change the output directory')
        parser_plot.add_argument(
            "-suffix",
            metavar="extra_info",
            type=str,
            default='',
            help='Add extra suffix if you used this in training')
        return parser_plot

    def make_parser_topology(self, parser_topology):
        parser_topology.add_argument(
            "-type",
            default='create_annovar_input', type=str,
            choices=['create_annovar_input', 'create_gene_network', 'create_pathway_KEGG', 'create_GTEx_network'],
            help="Create annovar input, create network topology from annovar output")
        parser_topology.add_argument(
            "-path",
            type=str,
            required=True,
            help="Path to the input data. For create_annovar_input this is the folder containing hase: genotype, "
                 "probes and individuals ")
        parser_topology.add_argument(
            '-study_name',
            type=str,
            required=True,
            help='Study name used in Convert. Name of the files in the genotype individuals and probe folders')
        parser_topology.add_argument(
            "-out",
            type=str,
            help="Path. Location of the results, default to ./processed_data/",
            default=os.getcwd() + '/processed_data/')
        return parser_topology



    def make_parser_interpret(self, parser_topology):
        parser_topology.add_argument(
            "-type",
            default='get_weight_scores', type=str,
            choices=['get_weight_scores', 'NID', 'RLIPP', 'DFIM',"PathExplain","DeepExplain"],
            help="choose interpretation method, choice")
        parser_topology.add_argument(
            "-resultpath",
            type=str,
            required=True,
            help="Path to the folder with the trained network (resultfolder) ")
        parser_topology.add_argument(
            '-layer',
            type=int,
            required=False,
            help='Select a layer for interpretation only necessary for NID')
        parser_topology.add_argument(
            '-num_eval',
            type=int,
            required=False,
            default = 100,
            help='Select the number of SNPs to eval in DFIM')
        parser_topology.add_argument(
            '-start_rank',
            type=int,
            required=False,
            default = 0,
            help='Multiprocessing, start from Nth ranked important variant')
        parser_topology.add_argument(
            '-end_rank',
            type=int,
            required=False,
            default = 0,
            help='Multiprocessing, stop at Nth ranked important SNP')
        parser_topology.add_argument(
            '-num_sample_pat',
            type=int,
            required=False,
            default = 1000,
            help='Select a number of patients to sample for DFIM')
        return parser_topology


if __name__ == '__main__':
    main()



import os
import sys

import numpy as np

from gennet.utils.hase.config import MAPPER_CHUNK_SIZE, basedir, CONVERTER_SPLIT_SIZE, PYTHON_PATH

os.environ['HASEDIR'] = basedir
if PYTHON_PATH is not None:
    for i in PYTHON_PATH: sys.path.insert(0, i)
from gennet.utils.hase.hdgwas.tools import Timer, Checker, study_indexes, Mapper, HaseAnalyser, merge_genotype, \
    check_np, check_converter
from gennet.utils.hase.hdgwas.converter import GenotypePLINK, GenotypeMINIMAC, GenotypeVCF
from gennet.utils.hase.hdgwas.data import Reader, MetaParData, MetaPhenotype
from gennet.utils.hase.hdgwas.fake import Encoder
from gennet.utils.hase.hdgwas.hdregression import HASE, A_inverse, B4
import argparse
import gc
from gennet.utils.hase.hdgwas.pard import partial_derivatives
from gennet.utils.hase.hdgwas.regression import haseregression
import pandas as pd
import time
from gennet.utils.hase.hdgwas.protocol import Protocol

__version__ = '1.1.0'

HEAD = "*********************************************************************\n"
HEAD += "* HASE: Framework for efficient high-dimensional association analyses \n"
HEAD += "* Version {V}\n".format(V=__version__)
HEAD += "* (C) 2015-2017 Gennady Roshchupkin and Hieab Adams\n"
HEAD += "* Erasmus MC, Rotterdam /  Department of Medical Informatics, Radiology and Epidemiology \n"
HEAD += "* GNU General Public License v3\n"
HEAD += "*********************************************************************\n"

if __name__ == '__main__':

    start = time.time()

    parser = argparse.ArgumentParser(description='Script to use HASE in command line')

    # MAIN SETTINGS
    parser.add_argument("-thr", type=float,
                        help="predefined threshold for t-value, if not defined save everything")

    parser.add_argument("-o", "--out", type=str, required=True, help="path to save result folder")

    parser.add_argument("-mode", required=True, type=str,
                        choices=['regression', 'converting', 'single-meta', 'meta-stage', 'encoding'],
                        help='Main argument, specify type of analysis'
                             '*****************'
                             'converting - before start to use any kind of analysis you need to convert data, it could take some time, '
                             'you need to do it ones and then everything will be super fast'
                             '*****************'
                             'single-meta - if you are the part of meta-analysis and will send data to main center, use this mode'
                             '*****************'
                             'encoding - if you are doing HD association, before send data to main center create fake data'
                             '*****************'
                             'meta-stage - if you are the main center of meta-analysis and already got precomputed data from all sites, use it'
                             '*****************'
                             'regression - if you want run everything yourself and do not use precomputed data use it'

                             'If you want to run single site full association analysis run (1) single-meta mode and then (2) meta-stage'

                             '*****************'
                        )

    parser.add_argument("-g", "--genotype", nargs='+', type=str, help="path/paths to genotype data folder")
    parser.add_argument("-ph", "--phenotype", nargs='+', type=str, help="path to phenotype data folder")
    parser.add_argument("-cov", "--covariates", type=str, help="path to covariates data folder")

    parser.add_argument("-derivatives", nargs='+', type=str, help="path to derivatives data folder")

    parser.add_argument("-protocol", type=str, help="path to study protocol file")

    parser.add_argument('-study_name', type=str, required=True, nargs='+',
                        help=' Name for saved genotype data, without ext')

    parser.add_argument('-mapper', type=str, help='Mapper data folder')

    parser.add_argument('-ref_name', type=str, default='1000Gp1v3_ref', help='Reference panel name')

    # ADDITIONAL SETTINGS
    parser.add_argument("-snp_id_inc", type=str, help="path to file with SNPs id to include to analysis")
    parser.add_argument("-snp_id_exc", type=str, help="path to file with SNPs id to exclude from analysis")
    parser.add_argument("-ph_id_inc", type=str, help="path to file with phenotype id to exclude from analysis")
    parser.add_argument("-ph_id_exc", type=str, help="path to file with phenotype id to exclude from analysis")

    # parser.add_argument("-ind_id_inc", type=str, help="path to file with individuals id to include to analysis") #TODO (low)
    # parser.add_argument("-ind_id_exc", type=str, help="path to file with individuals id to exclude from analysis")#TODO (low)
    # parser.add_argument("-cov_name_inc", type=str, help="path to file with covariates names to include to analysis")#TODO (low)
    # parser.add_argument("-cov_name_exc", type=str, help="path to file with covariates names to exclude from analysis")#TODO (low)

    parser.add_argument('-intercept', type=str, default='y', choices=['y', 'n'],
                        help='include intercept to regression, default yes')

    parser.add_argument('-maf', type=float, default=0.0, help='MAF for genetics data')

    parser.add_argument('-encoded', nargs='+', type=int, help='Value per every study, 1 - if encoded, 0 - if not')
    ###

    # FLAGS
    parser.add_argument('-hdf5', action='store_true', default=True, help='flag for genotype data format')
    parser.add_argument('-id', action='store_true', default=False,
                        help='Flag to convert minimac data to genotype per subject files first (default False)')
    parser.add_argument('-pd_full', action='store_true', default=False, help='For not HD association study')
    parser.add_argument('-effect_intercept', action='store_true', default=False,
                        help='Flag for add study effect to PD regression model')
    parser.add_argument('-permute_ph', action='store_true', default=False, help='Flag for phenotype permutation')
    parser.add_argument('-vcf', action='store_true', default=False, help='Flag for VCF data to convert')
    parser.add_argument('-np', action='store_true', default=True, help='Check BLAS/LAPACK/MKL')

    # TODO (low) save genotype after MAF
    ###

    ###CLUSTER SETTING
    parser.add_argument('-cluster', type=str, default='n', choices=['y', 'n'],
                        help=' Is it parallel cluster job, default no')
    parser.add_argument('-node', nargs='+', type=int, help='number of nodes / this node number, example: 10 2 ')
    ###

    # ADVANCED SETTINGS
    parser.add_argument('-interaction', type=str,
                        help='path to file with data for genotype interaction test')  # TODO (low)
    parser.add_argument('-mapper_chunk', type=int, help='Change mapper chunk size from config file')
    ###
    args = parser.parse_args()
    if not args.thr:
        print('WARNING!!! You did not set threshold for t-value, all results will be saved')
    if args.mapper_chunk:
        MAPPER_CHUNK_SIZE = args.mapper_chunk
    ARG_CHECKER = Checker()
    print(args)
    os.environ['HASEOUT'] = args.out

    if args.cluster == 'y':
        if args.node is not None:
            if args.node[1] > args.node[0]:
                raise ValueError('Node # {} > {} total number of nodes'.format(args.node[1], args.node[0]))

    if not os.path.isdir(args.out):
        print("Creating output folder {}".format(args.out))
        os.makedirs(args.out)

    if args.np:
        check_np()

    ################################### CONVERTING ##############################
    if args.mode == 'converting':

        # ARG_CHECKER.check(args,mode='converting')

        R = Reader('genotype')
        R.start(args.genotype[0], vcf=args.vcf)

        with Timer() as t:
            if R.format == 'PLINK':
                G = GenotypePLINK(args.study_name[0], reader=R)
                G.split_size = CONVERTER_SPLIT_SIZE
                G.plink2hdf5(out=args.out)

            elif R.format == 'MINIMAC':
                G = GenotypeMINIMAC(args.study_name[0], reader=R)
                if args.cluster == 'y':
                    G.cluster = True
                G.split_size = CONVERTER_SPLIT_SIZE
                G.MACH2hdf5(args.out, id=args.id)

            elif R.format == 'VCF':
                G = GenotypeVCF(args.study_name[0], reader=R)
                if args.cluster == 'y':
                    G.cluster = True
                G.split_size = CONVERTER_SPLIT_SIZE
                G.VCF2hdf5(args.out)
            else:
                raise ValueError('Genotype data should be in PLINK/MINIMAC/VCF format and alone in folder')

        check_converter(args.out, args.study_name[0])
        print(('Time to convert all data: {} sec'.format(t.secs)))

    ################################### ENCODING ##############################

    elif args.mode == 'encoding':

        # ARG_CHECKER.check(args,mode='encoding')
        mapper = Mapper()
        mapper.genotype_names = args.study_name
        mapper.chunk_size = MAPPER_CHUNK_SIZE
        mapper.reference_name = args.ref_name
        mapper.load_flip(args.mapper)
        mapper.load(args.mapper)

        phen = Reader('phenotype')
        phen.start(args.phenotype[0])

        gen = Reader('genotype')
        gen.start(args.genotype[0], hdf5=args.hdf5, study_name=args.study_name[0], ID=False)

        e = Encoder(args.out)
        e.study_name = args.study_name[0]

        row_index, ids = study_indexes(phenotype=phen.folder._data, genotype=gen.folder._data)
        with Timer() as t:

            e.matrix(len(ids), save=True)
            N_snps_read = 0
            while True:
                with Timer() as t_gen:
                    genotype = gen.get_next()
                    if isinstance(genotype, type(None)):
                        break

                    flip = mapper.flip[args.study_name[0]][N_snps_read:N_snps_read + genotype.shape[0]]
                    N_snps_read += genotype.shape[0]
                    flip_index = (flip == -1)
                    genotype = np.apply_along_axis(lambda x: flip * (x - 2 * flip_index), 0, genotype)
                    genotype = genotype[:, row_index[0]]
                    encode_genotype = e.encode(genotype, data_type='genotype')
                    e.save_hdf5(encode_genotype, os.path.join(args.out, 'encode_genotype'), info=gen.folder,
                                index=row_index[0])
                    encode_genotype = None
                    gc.collect()

                print(('Time to create fake genotype is {}sec'.format(t_gen.secs)))

            while True:
                with Timer() as t_phen:
                    phenotype = phen.get_next(index=row_index[1])
                    if isinstance(phenotype, type(None)):
                        break
                    encode_phenotype = e.encode(phenotype, data_type='phenotype')
                    if phen.folder.format == '.npy':
                        e.save_npy(encode_phenotype, save_path=os.path.join(args.out, 'encode_phenotype'),
                                   info=phen.folder, index=row_index[1])
                    if phen.folder.format in ['.csv', '.txt']:
                        e.save_csv(encode_phenotype, save_path=os.path.join(args.out, 'encode_phenotype'),
                                   info=phen.folder, index=row_index[1])
                    encode_phenotype = None
                    gc.collect()
            if phen.folder.format == '.npy':
                np.save(os.path.join(os.path.join(args.out, 'encode_phenotype', 'info_dic.npy')), e.phen_info_dic)
            print(('Time to create fake phenotype is {}sec'.format(t_phen.secs)))

        print(('Time to encode all data: {} sec'.format(t.secs)))

    ################################### SINGLE META STAGE ##############################

    elif args.mode == 'single-meta':

        # ARG_CHECKER.check(args,mode='single-meta')
        mapper = Mapper()
        mapper.genotype_names = args.study_name
        mapper.chunk_size = MAPPER_CHUNK_SIZE
        mapper.reference_name = args.ref_name
        mapper.load_flip(args.mapper)
        mapper.load(args.mapper)
        mapper.cluster = args.cluster
        mapper.node = args.node

        phen = Reader('phenotype')
        phen.start(args.phenotype[0])

        cov = Reader('covariates')
        cov.start(args.covariates)

        if cov.folder.n_files > 1:
            raise ValueError('In covariates folder should be only one file!')

        gen = Reader('genotype')
        gen.start(args.genotype[0], hdf5=args.hdf5, study_name=args.study_name[0], ID=False)

        with Timer() as t:
            partial_derivatives(save_path=args.out, COV=cov, PHEN=phen, GEN=gen,
                                MAP=mapper, MAF=args.maf, R2=None, B4_flag=args.pd_full,
                                study_name=args.study_name[0], intercept=args.intercept)
        print(('Time to compute partial derivatives : {} sec'.format(t.secs)))

    ################################### MULTI META STAGE ##############################

    elif args.mode == 'meta-stage':

        # ARG_CHECKER.check(args,mode='meta-stage')

        ##### Init data readers #####
        if args.derivatives is None:
            raise ValueError('For meta-stage analysis partial derivatives data are required!')
        mapper = Mapper()
        mapper.chunk_size = MAPPER_CHUNK_SIZE
        mapper.genotype_names = args.study_name
        mapper.reference_name = args.ref_name
        if args.snp_id_inc is not None:
            mapper.include = pd.DataFrame.from_csv(args.snp_id_inc, index_col=None)
            print('Include:')
            print(mapper.include.head())
            if 'ID' not in mapper.include.columns and (
                    'CHR' not in mapper.include.columns or 'bp' not in mapper.include.columns):
                raise ValueError('{} table does not have ID or CHR,bp columns'.format(args.snp_id_inc))
        if args.snp_id_exc is not None:
            mapper.exclude = pd.DataFrame.from_csv(args.snp_id_exc, index_col=None)
            print('Exclude:')
            print(mapper.exclude.head())
            if 'ID' not in mapper.exclude.columns and (
                    'CHR' not in mapper.exclude.columns or 'bp' not in mapper.exclude.columns):
                raise ValueError('{} table does not have ID or CHR,bp columns'.format(args.snp_id_exc))
        mapper.load(args.mapper)
        mapper.load_flip(args.mapper, encode=args.encoded)
        mapper.cluster = args.cluster
        mapper.node = args.node

        Analyser = HaseAnalyser()

        pard = []

        with Timer() as t:
            for i, j in enumerate(args.derivatives):
                pard.append(Reader('partial'))
                pard[i].start(j, study_name=args.study_name[i])
                pard[i].folder.load()

        print("time to set PD is {}s".format(t.secs))

        PD = [False if isinstance(i.folder._data.b4, type(None)) else True for i in pard]

        if np.sum(PD) != len(pard) and np.sum(PD) != 0:
            raise ValueError('All study should have b4 data for partial derivatives!')

        if args.protocol is not None:
            protocol = Protocol(args.protocol)
        else:
            protocol = None

        meta_pard = MetaParData(pard, args.study_name, protocol=protocol)

        if np.sum(PD) == 0:
            phen = []

            with Timer() as t:
                for i, j in enumerate(args.phenotype):
                    phen.append(Reader('phenotype'))
                    phen[i].start(j)
            print("Time to set pheno {} s".format(t.secs))
            meta_phen = MetaPhenotype(phen, include=args.ph_id_inc, exclude=args.ph_id_exc)

            N_studies = len(args.genotype)

            gen = []
            with Timer() as t:
                for i, j in enumerate(args.genotype):
                    gen.append(Reader('genotype'))
                    gen[i].start(j, hdf5=args.hdf5, study_name=args.study_name[i], ID=False)
            print("Time to set gen {}s".format(t.secs))

            row_index, ids = study_indexes(phenotype=tuple(i.folder._data for i in phen),
                                           genotype=tuple(i.folder._data for i in gen),
                                           covariates=tuple(i.folder._data.metadata for i in pard))
            if row_index[2].shape[0] != np.sum([i.folder._data.metadata['id'].shape[0] for i in pard]):
                raise ValueError(
                    'Partial Derivatives covariates have different number of subjects {} than genotype and phenotype {}'.format(
                        row_index[2].shape[0],
                        np.sum([i.folder._data.metadata['id'].shape[0] for i in pard])))
        while True:
            if mapper.cluster == 'n':
                SNPs_index, keys = mapper.get()
            else:
                ch = mapper.chunk_pop()
                if ch is None:
                    SNPs_index = None
                    break
                SNPs_index, keys = mapper.get(chunk_number=ch)

            if isinstance(SNPs_index, type(None)):
                break

            Analyser.rsid = keys
            if np.sum(PD) == 0:
                genotype = np.array([])
                with Timer() as t_g:
                    genotype = merge_genotype(gen, SNPs_index, mapper)
                    genotype = genotype[:, row_index[0]]
                print("Time to get G {}s".format(t_g.secs))
            # TODO (low) add interaction

            a_test = np.array([])
            b_cov = np.array([])
            C = np.array([])
            a_cov = np.array([])
            b4 = np.array([])

            if args.protocol is not None:
                if protocol.enable:
                    regression_model = protocol.regression_model()
            else:
                regression_model = None

            with Timer() as t_pd:
                if np.sum(PD) == 0:
                    a_test, b_cov, C, a_cov = meta_pard.get(SNPs_index=SNPs_index, regression_model=regression_model,
                                                            random_effect_intercept=args.effect_intercept)
                else:
                    a_test, b_cov, C, a_cov, b4 = meta_pard.get(SNPs_index=SNPs_index, B4=True,
                                                                regression_model=regression_model,
                                                                random_effect_intercept=args.effect_intercept)

            print("Time to get PD {}s".format(t_pd.secs))

            MAF = meta_pard.maf_pard(SNPs_index=SNPs_index)

            if args.maf != 0:
                filter = (MAF > args.maf) & (MAF < 1 - args.maf) & (MAF != 0.5)
                a_test = a_test[filter, :]
                Analyser.MAF = MAF[filter]
                if np.sum(PD) == 0:
                    genotype = genotype[filter, :]
                else:
                    b4 = b4[filter, :]
                Analyser.rsid = Analyser.rsid[filter]
                if a_test.shape[0] == 0:
                    print('NO SNPs > MAF')
                    continue
            else:
                Analyser.MAF = MAF

            a_inv = A_inverse(a_cov, a_test)
            N_con = a_inv.shape[1] - 1
            print('There are {} subjects in study.'.format(meta_pard.get_n_id()))
            DF = (meta_pard.get_n_id() - a_inv.shape[1])

            if np.sum(PD) == 0:

                while True:
                    phenotype = np.array([])
                    with Timer() as t_ph:
                        phenotype, phen_names = meta_phen.get()
                    print("Time to get PH {}s".format(t_ph.secs))

                    if isinstance(phenotype, type(None)):
                        meta_phen.processed = 0
                        print('All phenotypes processed!')
                        break
                    print(("Merged phenotype shape {}".format(phenotype.shape)))
                    # TODO (middle) select phen from protocol
                    phenotype = phenotype[row_index[1], :]
                    print(("Selected phenotype shape {}".format(phenotype.shape)))
                    keys = list(meta_pard.phen_mapper.dic.keys())
                    phen_ind_dic = {k: i for i, k in enumerate(keys)}
                    phen_ind = np.array([phen_ind_dic.get(i, -1) for i in phen_names])
                    if np.sum(phen_ind == -1) == len(phen_ind):
                        print('There is no common ids in phenotype files and PD data!')
                        break
                    else:
                        print('There are {} common ids in phenotype files and PD data!'.format(np.sum(phen_ind != -1)))
                    C_test = C[phen_ind]
                    b_cov_test = b_cov[:, phen_ind]

                    b4 = B4(phenotype, genotype)
                    print(("B4 shape is {}".format(b4.shape)))
                    t_stat, SE = HASE(b4, a_inv, b_cov_test, C_test, N_con, DF)

                    if mapper.cluster == 'y':
                        Analyser.cluster = True
                        Analyser.chunk = ch
                        Analyser.node = mapper.node[1]
                    Analyser.t_stat = t_stat
                    Analyser.SE = SE
                    Analyser.threshold = args.thr
                    Analyser.out = args.out
                    Analyser.save_result(phen_names[(phen_ind != -1)])

                    t_stat = None
                    Analyser.t_stat = None
                    gc.collect()

            else:
                t_stat, SE = HASE(b4, a_inv, b_cov, C, N_con, DF)

                Analyser.t_stat = t_stat
                Analyser.SE = SE
                Analyser.threshold = args.thr
                Analyser.out = args.out
                Analyser.save_result(np.array(list(meta_pard.phen_mapper.dic.keys())))

                t_stat = None
                Analyser.t_stat = None
                gc.collect()



    ################################### TO DO EVERYTHING IN ONE GO ##############################

    elif args.mode == 'regression':

        # ARG_CHECKER.check(args,mode='regression')

        print('START regression mode...')
        if args.mapper is None:
            os.environ['MAPPER'] = "False"  # need to set it here, before genotype Reader start

        phen = Reader('phenotype')
        phen.start(args.phenotype[0])
        phen.permutation = args.permute_ph

        cov = Reader('covariates')
        cov.start(args.covariates)

        interaction = None
        if args.interaction is not None:
            interaction = Reader('interaction')
            interaction.start(args.interaction)

        if (cov.folder.n_files > 1 and cov.folder.format != '.npy') or (
                cov.folder.n_files > 2 and cov.folder.format == '.npy'):  # TODO (middle) test
            raise ValueError('In covariates folder should be only one file!')

        gen = []
        for i, j in enumerate(args.genotype):
            gen.append(Reader('genotype'))
            gen[i].start(j, hdf5=args.hdf5, study_name=args.study_name[i], ID=False)

        if args.mapper is not None:
            mapper = Mapper()
            mapper.chunk_size = MAPPER_CHUNK_SIZE
            mapper.genotype_names = args.study_name
            mapper.reference_name = args.ref_name
            if args.snp_id_inc is not None:
                mapper.include = pd.DataFrame.from_csv(args.snp_id_inc, index_col=None)
                print('Include:')
                print(mapper.include.head())
                if 'ID' not in mapper.include.columns and (
                        'CHR' not in mapper.include.columns or 'bp' not in mapper.include.columns):
                    raise ValueError('{} table does not have ID or CHR,bp columns'.format(args.snp_id_inc))
            if args.snp_id_exc is not None:
                mapper.exclude = pd.DataFrame.from_csv(args.snp_id_exc, index_col=None)
                print('Exclude:')
                print(mapper.exclude.head())
                if 'ID' not in mapper.exclude.columns and (
                        'CHR' not in mapper.exclude.columns or 'bp' not in mapper.exclude.columns):
                    raise ValueError('{} table does not have ID or CHR,bp columns'.format(args.snp_id_exc))
            mapper.load(args.mapper)
            mapper.load_flip(args.mapper)
            mapper.cluster = args.cluster
            mapper.node = args.node
        else:
            if len(args.genotype) == 1:
                mapper = Mapper()
                mapper.chunk_size = MAPPER_CHUNK_SIZE
                mapper.genotype_names = args.study_name
                mapper.reference_name = args.ref_name
                mapper.cluster = args.cluster
                mapper.node = args.node
                mapper.n_study = 1
                mapper.n_keys = gen[0].folder._data.names.shape[0]
                mapper.keys = np.array(gen[0].folder._data.names.tolist())
                mapper.values = np.array(list(range(mapper.n_keys))).reshape(-1, 1)
                mapper.flip[args.study_name[0]] = np.array([1] * mapper.n_keys)
                if args.snp_id_exc is not None or args.snp_id_inc is not None:
                    raise ValueError('You can not exclude or include variants to analysis without mapper!')
            else:
                raise ValueError('You can not run regression analysis with several genotype data without mapper!')
        # mapper=None

        Analyser = HaseAnalyser()
        Analyser.threshold = args.thr
        Analyser.out = args.out
        haseregression(phen, gen, cov, mapper, Analyser, args.maf, intercept=args.intercept, interaction=interaction)

end = time.time()

print(('experiment finished in {} s'.format((end - start))))



import gc
import os
import shutil
import signal
import subprocess

import numpy as np
import pandas as pd
import tables

from gennet.utils.hase.hdgwas.data import MINIMACHDF5Folder
from gennet.utils.hase.hdgwas.tools import Timer


class Genotype(object):
    def __init__(self):
        self.file_name = None
        self.reader = None
        self.probes = None
        self.individuals = None
        self.genotype = None
        self.out = None


class GenotypeHDF5(Genotype):
    def __init__(self, name, force=True):
        super(GenotypeHDF5, self).__init__()
        self.h5_name = '%s.h5' % name
        self.file_name = name
        self.pytable_filters = tables.Filters(complevel=9, complib='zlib')
        self.h5_gen_file = None
        self.h5_ind_file = None
        self.h5_pr_file = None
        self.gen_iter = 0

    def write_data(self, type, overwrite=True):

        type_dic = {'gen': ['genotype', self.h5_gen_file],
                    'ind': ['individuals', self.h5_ind_file],
                    'pr': ['probes', self.h5_pr_file]}

        if (not overwrite) and os.path.isfile(os.path.join(self.out, type_dic[type][0], self.h5_name)):
            print(('File %s found. Please remove manually.' % self.h5_name))
            return
        else:
            if type == 'pr':
                self.h5_pr_file = tables.open_file(os.path.join(self.out, type_dic[type][0], self.h5_name), 'w',
                                                   title=self.file_name)
                self.h5_pr_file.close()  # need to close file before join data
            elif type == 'ind':
                self.h5_ind_file = tables.open_file(os.path.join(self.out, type_dic[type][0], self.h5_name), 'w',
                                                    title=self.file_name)
            elif type == 'gen':
                self.h5_gen_file = tables.open_file(
                    os.path.join(self.out, type_dic[type][0], str(self.gen_iter) + '_' + self.h5_name),
                    'w', title=self.file_name)
                self.gen_iter += 1

    def close(self):
        self.h5_gen_file.close()
        self.h5_ind_file.close()
        self.h5_pr_file.close()

    def summary(self):
        raise (NotImplementedError)


class GenotypePLINK(GenotypeHDF5):

    def __init__(self, name, reader=None):
        super(GenotypePLINK, self).__init__(name)
        self.reader = reader
        self.split_size = None

    def convert_individuals(self):

        individuals = self.reader.folder.get_fam()

        self.h5_ind_file.create_table(self.h5_ind_file.root, 'individuals', individuals,
                                      title='Individuals', filters=self.pytable_filters)
        self.h5_ind_file.root.individuals[:] = individuals
        self.individuals = self.h5_ind_file.root.individuals[:]
        self.n_ind = len(individuals)

    # @profile
    def convert_probes(self, chunk_size=100000):

        if os.path.isfile(os.path.join(self.out, 'probes', self.h5_name)):
            os.remove(os.path.join(self.out, 'probes', self.h5_name))
        hash_table = {'keys': np.array([], dtype=int), 'allele': np.array([])}
        i = 0
        chunk = np.array([])
        while True:
            chunk = self.reader.folder.get_bim(chunk_size)
            if isinstance(chunk, type(None)):
                break
            chunk.columns = ['CHR', 'ID', 'distance', 'bp', 'allele1', 'allele2']
            hash_1 = chunk.allele1.apply(hash)
            hash_2 = chunk.allele2.apply(hash)
            k, indices = np.unique(np.append(hash_1, hash_2), return_index=True)
            s = np.append(chunk.allele1, chunk.allele2)[indices]
            ind = np.invert(np.in1d(k, hash_table['keys']))
            hash_table['keys'] = np.append(hash_table['keys'], k[ind])
            hash_table['allele'] = np.append(hash_table['allele'], s[ind])
            chunk.allele1 = hash_1
            chunk.allele2 = hash_2
            # WARNING!!! doesn't work on windows
            chunk.to_hdf(os.path.join(self.out, 'probes', self.h5_name), key='probes', format='table',
                         data_columns=True, append=True,
                         complib='zlib', complevel=9, min_itemsize=45)
            gc.collect()
            i += 1
        pd.DataFrame.from_dict(hash_table).to_csv(
            os.path.join(self.out, 'probes', self.file_name + '_hash_table.csv.gz'), index=False, compression='gzip',
            sep='\t')
        print(('Number of Probes {} converted'.format(self.reader.folder.N_probes)))

    # @profile
    def convert_genotypes(self):

        chunk_size = self.split_size
        if chunk_size is None:
            raise ValueError('CONVERTER_SPLIT_SIZE does not define in config file!')
        G = np.array([])
        # self.reader.folder.processed=0
        while True:
            with Timer() as t:
                G = self.reader.folder.get_bed(chunk_size)
                if isinstance(G, type(None)):
                    break

            print(('Time to read {} SNPs is {} s'.format(G.shape[0], t.secs)))

            self.write_data('gen')
            atom = tables.Int8Atom()
            self.genotype = self.h5_gen_file.create_carray(self.h5_gen_file.root, 'genotype', atom,
                                                           (G.shape),
                                                           title='Genotype', filters=self.pytable_filters)
            with Timer() as t:
                self.genotype[:] = G

            print(('Time to write {} SNPs is {} s'.format(G.shape[0], t.secs)))

            self.h5_gen_file.close()
            G = None
            gc.collect()

    def plink2hdf5(self, out, force=True):

        if force:
            try:
                os.makedirs(os.path.join(out, 'genotype'))
                os.makedirs(os.path.join(out, 'individuals'))
                os.makedirs(os.path.join(out, 'probes'))
            except:
                print(('Directories "genotype","probes","individuals" are already exist in {}...'.format(self.out)))

        self.out = out
        self.write_data('ind')
        self.convert_individuals()
        self.h5_ind_file.close()

        self.reader.folder.processed = 0
        self.write_data('pr')
        self.convert_probes()

        self.reader.folder.processed = 0
        self.convert_genotypes()

    def _summary(self, head=10):
        pass  # TODO (low) rewrite to get statistic not load to memory

        print(('Number of Probes: %d' % None))
        print(('Number of Individuals: %d' % None))
        print(('The genotype matrix is of size %d by %d' % None))


class GenotypeMINIMAC(object):

    def __init__(self, name, reader=None):
        self.reader = reader
        self.study_name = name
        self.split_size = None
        self.hdf5_iter = 0
        self.pytable_filter = tables.Filters(complevel=9, complib='zlib')
        self.cluster = False

    def save_hdf5_chunk(self, data, out, name):
        print('Saving chunk...{}'.format(os.path.join(out, 'genotype', str(self.hdf5_iter) + '_' + name + '.h5')))
        h5_gen_file = tables.open_file(
            os.path.join(out, 'genotype', str(self.hdf5_iter) + '_' + name + '.h5'), 'w', title=name)

        atom = tables.Float16Atom()  # TODO (low) check data format
        genotype = h5_gen_file.create_carray(h5_gen_file.root, 'genotype', atom,
                                             (data.shape),
                                             title='Genotype',
                                             filters=self.pytable_filter)
        genotype[:] = data
        h5_gen_file.close()
        genotype = None
        data = None
        gc.collect()
        self.hdf5_iter += 1

    def MACH2hdf5(self, out, id=False, remove_id=False):
        FNULL = open(os.devnull, 'w')
        subprocess.call(['bash', os.path.join(os.environ['HASEDIR'], 'tools', 'minimac2hdf5.sh'),
                         self.reader.folder.path, out, os.environ['HASEDIR'], self.study_name], shell=False,
                        stderr=subprocess.STDOUT)

        if id:
            if self.cluster:
                ind = pd.read_hdf(os.path.join(out, 'individuals', self.study_name + '.h5'), 'individuals').individual
                N = ind.shape[0]
                print('Submit to cluster!')
                cmd = "qsub -sync y -t 1-{} {} {}".format(N, os.path.join(os.environ['HASEDIR'], 'tools',
                                                                          'qsub_helper.sh'),
                                                          os.path.join(out, 'id_convert.sh'))
                print(cmd)
                proc = subprocess.Popen(cmd, shell=True, stderr=subprocess.STDOUT, stdout=subprocess.PIPE).communicate()
            else:
                proc = subprocess.Popen(['bash', os.path.join(out, 'id_convert.sh')], shell=False,
                                        stderr=subprocess.STDOUT)
                print(proc.communicate())
            self.folder = MINIMACHDF5Folder(out, self.study_name)
            self.folder.pool.split_size = self.split_size
            self.folder.pool.chunk_size = self.split_size
            print('Start to convert id files to chunk files...')
            while True:
                data = self.folder.get_next()
                if data is None:
                    break
                self.save_hdf5_chunk(data, out, self.study_name)
                gc.collect()

            print('Finished')
            if remove_id:
                self.folder.pool.remove(type='all')
            else:
                try:
                    os.makedirs(os.path.join(out, 'id_genotype'))
                except:
                    print('id_genotype folder already exist')

                self.folder.pool.move(os.path.join(out, 'id_genotype'), type='all')

            shutil.move(os.path.join(out, 'id_convert.sh'), os.path.join(out, 'tmp_files', 'id_convert.sh'))
            shutil.move(os.path.join(out, 'SUB_FAM.txt'), os.path.join(out, 'tmp_files', 'SUB_FAM.txt'))
            shutil.move(os.path.join(out, 'SUB_ID.txt'), os.path.join(out, 'tmp_files', 'SUB_ID.txt'))

        else:
            f = open(os.path.join(out, 'minimac_convert.sh'), 'w')
            probes = pd.HDFStore(os.path.join(out, 'probes', self.study_name + '.h5'), 'r')
            N_probes = probes.get_storer('probes').nrows
            print('There are {} probes'.format(N_probes))
            chunk = np.vstack(
                ((np.arange(0, N_probes, self.split_size) + 1)[:-1], np.arange(0, N_probes, self.split_size)[1:]))
            N_jobs = chunk.shape[1]
            ch = [0, 0]
            i_ch = -1
            for i_ch in range(chunk.shape[1]):
                ch = chunk[:, i_ch]
                # print ch
                l = 'bash {} {} {} {} {} {} {} {} \n'.format(
                    os.path.join(os.environ['HASEDIR'], 'tools', 'minimacGenotype2hdf5.sh'),
                    self.reader.folder.path,
                    out,
                    os.environ['HASEDIR'],
                    self.study_name,
                    ch[0],
                    ch[1],
                    i_ch
                )
                f.write(l)
            if ch[1] != N_probes:
                l = 'bash {} {} {} {} {} {} {} {} \n'.format(
                    os.path.join(os.environ['HASEDIR'], 'tools', 'minimacGenotype2hdf5.sh'),
                    self.reader.folder.path,
                    out,
                    os.environ['HASEDIR'],
                    self.study_name,
                    ch[1] + 1,
                    N_probes,
                    i_ch + 1
                )
                f.write(l)
                N_jobs += 1
            f.close()
            if self.cluster:
                print('Submit to cluster!')
                cmd = "qsub -sync y -t 1-{} {} {}".format(N_jobs, os.path.join(os.environ['HASEDIR'], 'tools',
                                                                               'qsub_helper.sh'),
                                                          os.path.join(out, 'minimac_convert.sh'))
                print(cmd)
                proc = subprocess.Popen(cmd, shell=True, stderr=subprocess.STDOUT, stdout=subprocess.PIPE).communicate()
            else:
                proc = subprocess.Popen(['bash', os.path.join(out, 'minimac_convert.sh')], shell=False, stderr=FNULL)
                print(proc.communicate())

            shutil.move(os.path.join(out, 'minimac_convert.sh'), os.path.join(out, 'tmp_files', 'minimac_convert.sh'))
            shutil.move(os.path.join(out, 'id_convert.sh'), os.path.join(out, 'tmp_files', 'id_convert.sh'))
            shutil.move(os.path.join(out, 'SUB_FAM.txt'), os.path.join(out, 'tmp_files', 'SUB_FAM.txt'))
            shutil.move(os.path.join(out, 'SUB_ID.txt'), os.path.join(out, 'tmp_files', 'SUB_ID.txt'))

        shutil.move(os.path.join(out, 'files_order.txt'), os.path.join(out, 'tmp_files', 'files_order.txt'))
        shutil.move(os.path.join(out, 'info.txt'), os.path.join(out, 'tmp_files', 'info.txt'))

    def summary(self):
        pass


class GenotypeVCF(object):

    def __init__(self, name, reader=None):
        self.reader = reader
        self.study_name = name
        self.split_size = None
        self.hdf5_iter = 0
        self.pytable_filter = tables.Filters(complevel=9, complib='zlib')
        self.cluster = False

    def VCF2hdf5(self, out):

        FNULL = open(os.devnull, 'w')
        subprocess.call(['bash', os.path.join(os.environ['HASEDIR'], 'tools', 'VCF2hdf5.sh'),
                         self.reader.folder.path, out, os.environ['HASEDIR'], self.study_name], shell=False,
                        stderr=subprocess.STDOUT,
                        preexec_fn=lambda: signal.signal(signal.SIGPIPE, signal.SIG_DFL)
                        )

        f = open(os.path.join(out, 'vcf_convert.sh'), 'w')
        probes = pd.HDFStore(os.path.join(out, 'probes', self.study_name + '.h5'), 'r')
        N_probes = probes.get_storer('probes').nrows
        print('There are {} probes'.format(N_probes))
        chunk = np.vstack(
            ((np.arange(0, N_probes, self.split_size) + 1)[:-1], np.arange(0, N_probes, self.split_size)[1:]))
        N_jobs = chunk.shape[1]
        ch = [0, 0]
        i_ch = -1
        for i_ch in range(chunk.shape[1]):
            ch = chunk[:, i_ch]
            # print ch
            l = 'bash {} {} {} {} {} {} {} {} \n'.format(
                os.path.join(os.environ['HASEDIR'], 'tools', 'VCFGenotype2hdf5.sh'),
                self.reader.folder.path,
                out,
                os.environ['HASEDIR'],
                self.study_name,
                ch[0],
                ch[1],
                i_ch
            )
            f.write(l)
        if ch[1] != N_probes:
            l = 'bash {} {} {} {} {} {} {} {} \n'.format(
                os.path.join(os.environ['HASEDIR'], 'tools', 'VCFGenotype2hdf5.sh'),
                self.reader.folder.path,
                out,
                os.environ['HASEDIR'],
                self.study_name,
                ch[1] + 1,
                N_probes,
                i_ch + 1
            )
            f.write(l)
            N_jobs += 1
        f.close()
        if self.cluster:
            print('Submit to cluster!')
            cmd = "qsub -sync y -t 1-{} {} {}".format(N_jobs,
                                                      os.path.join(os.environ['HASEDIR'], 'tools', 'qsub_helper.sh'),
                                                      os.path.join(out, 'vcf_convert.sh'))
            print(cmd)
            proc = subprocess.Popen(cmd, shell=True, stderr=subprocess.STDOUT, stdout=subprocess.PIPE).communicate()
        else:
            proc = subprocess.Popen(['bash', os.path.join(out, 'vcf_convert.sh')], shell=False, stderr=FNULL)
            print(proc.communicate())

        shutil.move(os.path.join(out, 'vcf_convert.sh'), os.path.join(out, 'tmp_files', 'vcf_convert.sh'))
        shutil.move(os.path.join(out, 'SUB_ID.txt'), os.path.join(out, 'tmp_files', 'SUB_ID.txt'))
        shutil.move(os.path.join(out, 'snps_count.txt'), os.path.join(out, 'tmp_files', 'snps_count.txt'))

        shutil.move(os.path.join(out, 'files_order.txt'), os.path.join(out, 'tmp_files', 'files_order.txt'))
        shutil.move(os.path.join(out, 'info.txt'), os.path.join(out, 'tmp_files', 'info.txt'))
